/*
 * (C) Copyright 2011
 * CompuLab, Ltd., <www.compulab.co.il>
 * Ilya Ledvich <ilya@compulab.co.il>
 * Igor Grinberg <grinberg@compulab.co.il>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc.
 */

#include <common.h>

#include <asm/arch/cpu.h>
#include <asm/arch/bits.h>
#include <asm/arch/sys_info.h>
#include <asm/arch/mem.h>

#include "common.h"
#include "board.h"

/*
 * TODO: Move it to common place so that should be used
 * for all OMAP series of devices
 */
u32 is_cpu_family(void)
{
	u32 cpuid = 0, cpu_family = 0;
	u16 hawkeye;

	__asm__ __volatile__("mrc p15, 0, %0, c0, c0, 0":"=r"(cpuid));
	if ((cpuid & 0xf) == 0x0) {
		cpu_family = CPU_OMAP34XX;
	} else {
		cpuid = __raw_readl(OMAP34XX_CONTROL_ID);
		hawkeye = (cpuid >> HAWKEYE_SHIFT) & 0xffff;

		switch (hawkeye) {
			case HAWKEYE_AM35XX:
				cpu_family = CPU_AM35XX;
				break;
			case HAWKEYE_OMAP36XX:
				cpu_family = CPU_OMAP36XX;
				break;
			case HAWKEYE_OMAP34XX:
			default:
				cpu_family = CPU_OMAP34XX;
		}
	}

	return cpu_family;
}

/*****************************************************************************
 * cpu_has_iva() - check if CPU has imaging video and audio (IVA) accelerator
 *****************************************************************************/
int cpu_has_iva(void)
{
	if (cpu_is_am35xx())
		return 0;

	return 1;
}

/************************************************************
 * get_device_type(): tell if GP/HS/EMU/TST
 ************************************************************/
u32 get_device_type(void)
{
	return (__raw_readl(CONTROL_STATUS) & DEVICE_MASK) >> 8;
}

/************************************************
 * get_sysboot_value(void) - return SYS_BOOT[5:0]
 ************************************************/
u32 get_sysboot_value(void)
{
	return __raw_readl(CONTROL_STATUS) & SYSBOOT_MASK;
}

/******************************************
 * get_cpu_rev(void) - extract version info
 ******************************************/
u32 get_cpu_rev(void)
{
	u32 cpuid = 0;

	/*
	 * On ES1.0 the IDCODE register is not exposed on L4
	 * so using CPU ID to differentiate
	 * between ES2.0 and ES1.0.
	 */
	__asm__ __volatile__("mrc p15, 0, %0, c0, c0, 0":"=r" (cpuid));
	if ((cpuid & 0xf) == 0x0)
		return CPU_3430_ES1;
	else
		return CPU_3430_ES2;
}

/****************************************************************
 * sr32 - clear & set a value in a bit range for a 32 bit address
 ****************************************************************/
void sr32(u32 addr, u32 start_bit, u32 num_bits, u32 value)
{
	u32 tmp, msk = 0;

	msk = 1 << num_bits;
	--msk;
	tmp = __raw_readl(addr) & ~(msk << start_bit);
	tmp |= value << start_bit;
	__raw_writel(tmp, addr);
}

/********************************************************************
 * wait_on_value() - common routine to allow waiting for changes in
 *   volatile regs.
 ********************************************************************/
u32 wait_on_value(u32 read_bit_mask, u32 match_value, u32 read_addr, u32 bound)
{
	u32 i = 0, val;

	do {
		++i;
		val = __raw_readl(read_addr) & read_bit_mask;
		if (val == match_value)
			return 1;
		if (i == bound)
			return 0;
	} while (1);
}

/* Permission values for registers - Full fledged permissions to all */
#define UNLOCK_1		0xFFFFFFFF
#define UNLOCK_2		0x00000000
#define UNLOCK_3		0x0000FFFF

/***************************************************
 * Routine: secure_unlock
 * Description: Setup security registers for access
 * (GP Device only)
 ***************************************************/
static void secure_unlock(void)
{
	/* Protection Module Register Target APE (PM_RT)*/
	__raw_writel(UNLOCK_1, RT_REQ_INFO_PERMISSION_1);
	__raw_writel(UNLOCK_1, RT_READ_PERMISSION_0);
	__raw_writel(UNLOCK_1, RT_WRITE_PERMISSION_0);
	__raw_writel(UNLOCK_2, RT_ADDR_MATCH_1);

	__raw_writel(UNLOCK_3, GPMC_REQ_INFO_PERMISSION_0);
	__raw_writel(UNLOCK_3, GPMC_READ_PERMISSION_0);
	__raw_writel(UNLOCK_3, GPMC_WRITE_PERMISSION_0);

	__raw_writel(UNLOCK_3, OCM_REQ_INFO_PERMISSION_0);
	__raw_writel(UNLOCK_3, OCM_READ_PERMISSION_0);
	__raw_writel(UNLOCK_3, OCM_WRITE_PERMISSION_0);
	__raw_writel(UNLOCK_2, OCM_ADDR_MATCH_2);

	if (cpu_has_iva()) {
		/* IVA Changes */
		__raw_writel(UNLOCK_3, IVA2_REQ_INFO_PERMISSION_0);
		__raw_writel(UNLOCK_3, IVA2_READ_PERMISSION_0);
		__raw_writel(UNLOCK_3, IVA2_WRITE_PERMISSION_0);
	}

	__raw_writel(UNLOCK_1, SMS_RG_ATT0); /* SDRC region 0 public */
}

/*********************************************************
 * Routine: try_unlock_memory()
 * Description: If chip is GP type, unlock the SRAM for
 *  general use.
 *********************************************************/
void try_unlock_memory(void)
{
	/*
	 * if GP device unlock device SRAM for general use
	 * secure code breaks for Secure/Emulation device - HS/E/T
	 */
	if (get_device_type() == GP_DEVICE)
		secure_unlock();
}

/*****************************************************
 * Routine: wait_for_command_complete
 * Description: Wait for posting to finish on watchdog
 *****************************************************/
static void wait_for_command_complete(unsigned int wd_base)
{
	int pending = 1;

	do {
		pending = __raw_readl(wd_base + WWPS);
	} while (pending);
}

/***************************************
 * Routine: watchdog_init
 * Description: Shut down watch dogs
 ***************************************/
void watchdog_init(void)
{
	/*
	 * There are 3 watch dogs WD1=Secure, WD2=MPU, WD3=IVA. WD1 is
	 * either taken care of by ROM (HS/EMU) or not accessible (GP).
	 * We need to take care of WD2-MPU or take a PRCM reset.  WD3
	 * should not be running and does not generate a PRCM reset.
	 */
	sr32(CM_FCLKEN_WKUP, 5, 1, 1);
	sr32(CM_ICLKEN_WKUP, 5, 1, 1);
	wait_on_value(BIT5, 0x20, CM_IDLEST_WKUP, 5); /* some issue here */

	__raw_writel(WD_UNLOCK1, WD2_BASE + WSPR);
	wait_for_command_complete(WD2_BASE);
	__raw_writel(WD_UNLOCK2, WD2_BASE + WSPR);
}

#ifdef DEBUG_LED
static u32 led_gpio_base(void)
{
	switch (DEBUG_LED_GPIO_BANK) {
	case 1:
		return OMAP34XX_GPIO1_BASE;
	case 2:
		return OMAP34XX_GPIO2_BASE;
	case 3:
		return OMAP34XX_GPIO3_BASE;
	case 4:
		return OMAP34XX_GPIO4_BASE;
	case 5:
		return OMAP34XX_GPIO5_BASE;
	case 6:
	default:;
	}

	return OMAP34XX_GPIO6_BASE;
}

void led_enable(void)
{
	/* Set DEBUG LED GPIO pin as output */
	sr32(led_gpio_base() + 0x34, DEBUG_LED_GPIO_OFFSET, 1, 0);
}

/*********************************************************
 * Routine: led_blink
 * Description: Blinks debug LED
 * Use fast=1 for error indication, fast=0 for debug
 *********************************************************/
void led_blink(int times, int fast)
{
	int i;
	unsigned int timeout = fast ? 10000000 : 30000000;

	for (i = 0; i < times; i++) {
		/* turn the LED on */
		__raw_writel(1 << DEBUG_LED_GPIO_OFFSET,
			     led_gpio_base() + 0x94);
		delay(timeout);

		/* turn the LED off */
		__raw_writel(1 << DEBUG_LED_GPIO_OFFSET,
			     led_gpio_base() + 0x90);
		delay(timeout);
	}
}
#endif /* DEBUG_LED */
